"""
Source:
    https://en.wikipedia.org/wiki/Mersenne_Twister#Python_implementation
"""

import math


def _int32(x):
    # Get the 32 least significant bits.
    return int(0xFFFFFFFF & x)

class MT19937:

    def __init__(self, seed):
        # Initialize the index to 0
        self.index = 624
        self.mt = [0] * 624
        self.mt[0] = seed  # Initialize the initial state to the seed
        for i in range(1, 624):
            self.mt[i] = _int32(
                1812433253 * (self.mt[i - 1] ^ (self.mt[i - 1] >> 30)) + i)

    def extract_number(self):
        if self.index >= 624:
            self.twist()

        y = self.mt[self.index]

        # Right shift by 11 bits
        y = y ^ y >> 11
        # Shift y left by 7 and take the bitwise and of 2636928640
        y = y ^ y << 7 & 2636928640
        # Shift y left by 15 and take the bitwise and of y and 4022730752
        y = y ^ y << 15 & 4022730752
        # Right shift by 18 bits
        y = y ^ y >> 18

        self.index = self.index + 1

        return _int32(y)

    def twist(self):
        for i in range(624):
            # Get the most significant bit and add it to the less significant
            # bits of the next number
            y = _int32((self.mt[i] & 0x80000000) +
                       (self.mt[(i + 1) % 624] & 0x7fffffff))
            self.mt[i] = self.mt[(i + 397) % 624] ^ y >> 1

            if y % 2 != 0:
                self.mt[i] = self.mt[i] ^ 0x9908b0df
        self.index = 0
    
    
    def rand_int(self, endpoint1, endpoint2):
        """Get an integer between the two endpoints."""
        
        # Organize the input.
        low = min(endpoint1, endpoint2)
        high = max(endpoint1, endpoint2)
        diff = high - low + 1
        
        # Get a random number.
        r = self.extract_number()
        
        # Transform it.
        mult = diff / 0xFFFFFFFF
        return min(math.floor((r * mult) + low), high)
    
    
    def rand(self):
        """Get a 0 to 1 float."""
        return self.extract_number()


import base64
import binascii
from bitarray import bitarray
import hashlib
import os

from .exceptions import ConfigError
from .utils import (
        base64_to_bitarray, bitarray_to_base64, bitarray_to_bytes,
        bitarray_to_hex, bitarray_to_int, bitarray_to_str, bytes_to_bitarray,
        hex_to_bitarray, int_to_bitarray, str_to_bitarray)


class BitCollection:
    """Holds, manipulates, and expresses bitarrays in various formats.
    
    A BitCollection is an abstraction from bitarray, which only supports
    a few interfaces for data. This abstraction focuses on a variety of
    input and output formats, as well as the ability to sew in bits from
    each of these formats.
    """
    
    def __init__(self, b=None):
        """Make a new collection out of a bytearray.
        
        Args:
            b (bitarray): Source to create collection from.
        
        """
        # Setup the parameters
        if b:
            if not isinstance(b, bitarray):
                raise ConfigError('b must be a bitearray')
            self.content = b
        else:
            self.content = bitarray()
    
    
    @classmethod
    def from_base64(cls, s, url_safe=False):
        """Creates a new collection from a base64 string.
        
        Args:
            s (str): A base64 string to ingest to the collection.
            url_safe (Optional[bool]): If true, substitute '-_' with
                '+/'.
        
        Return:
            BitCollection: new instance.
            
        """
        return cls(base64_to_bitarray(s, url_safe=url_safe))
    
    
    @classmethod
    def from_bytes(cls, b):
        """Creates a new collection from a hexidecimal string.
        
        Args:
            b (bytes): Bytes to start a collection with.
        
        Returns:
            BitCollection: new instance.
        
        """
        return cls(bytes_to_bitarray(b))
    
    
    @classmethod
    def from_hex(cls, s):
        """Creates a new collection from a hexidecimal string.
        
        Args:
            s (str): A hexidecimal string to ingest to the
                collection.
        
        Returns:
            BitCollection: new instance.
        
        Raises:
            ValueError: s isn't a valid hexidecimal number.
        
        """
        return cls(hex_to_bitarray(s))
    
    
    @classmethod
    def from_int(cls, i, bits):
        """Creates a new collection from a base64 string.
        
        Args:
            i (int): An integer to encode into bits.
            bits (int): Number of bits to assign this integer to.
        
        Returns:
            BitCollection: new instance.
            
        """
        return cls(int_to_bitarray(i, bits=bits))
    
    
    @classmethod
    def from_random(cls, bits):
        """Generates a totally random BitCollection.
        
        Args:
            bits (int): Length of the BitCollection.
        
        Return:
            BitCollection: new instance.
        
        """
        # The source length needs to be divisible by 8 and encompassing.
        mod = bits % 8
        if mod:
            bits += (8 - mod)
        
        # Generate off the os's better randomness.
        bytes_ = os.urandom(bits // 8)
        
        # Create BitCollection from resulting bytes.
        b = cls.from_bytes(bytes_)
        
        # Shave off undwanted bits.
        if mod:
            for i in range(8 - mod):
                b.pop()
        
        # Return the BitCollection
        return b
    
    
    def extract(self, positions):
        """Extract BitCollection from this collection.
        
        Args:
            positions (list): The positions to extract bits from the
                collection.
        
        Returns:
            BitCollection: bits in the order of extraction.
        
        """
        return self.__class__(self.extract_bitarray(positions=positions))
    
    
    def extract_bitarray(self, positions):
        """Extract bits from this collection at the given positions.
        
        Args:
            positions (list): The positions to extract bits from the
                collection.
        
        Returns:
            bitarray: bits in the order of extraction.
        
        """
        # Roll through each position.
        bits = bitarray()
        for j, position in enumerate(positions):
            
            # Put the value into this collection
            bits.append(self.content.pop(position))
        
        # Return reorder the bits and return
        bits.reverse()
        return bits
    
    
    def extract_bytes(self, positions):
        """Extract bytes from this collection at the given positions.
        
        Args:
            positions (list): The positions to extract bits from the
                collection.
        
        Returns:
            bytes: in the order of extraction.
        
        """
        # Extract the bits
        bits = self.extract_bitarray(positions=positions)
        
        # Convert to bytes
        return bitarray_to_bytes(bits)
    
    
    def extract_int(self, positions):
        """Extract the bits of the integer from this collection.
        
        Args:
            positions (list): The positions to extract bits from the
                collection.
        
        Returns:
            int: in the order of extraction.
        
        """
        # Roll through each position.
        output = 0
        total = len(positions) - 1
        for j, position in enumerate(positions):
            
            # Get the bit from this collection
            bit = self.content.pop(position)
            
            # Add to the output
            output |= bit << (total - j)
        
        # Return the reconstructed integer
        return output
    
    
    def extract_hex(self, positions):
        """Extract a hexadecimal string from this collection.
        
        Args:
            positions (list): The positions to extract bits from the
                collection.
        
        Returns:
            str: hexadecimal string in order of extraction.
        
        """
        # Extract the bits
        bits = self.extract_bitarray(positions=positions)
        
        # Convert to hex
        return bitarray_to_hex(bits)
    
    
    def extract_base64(self, positions, url_safe=False):
        """Extract bits from this collection at the given positions.
        
        Args:
            positions (list): The positions to extract bits from the
                collection.
            url_safe (Optional[bool]): Whether the output string
                contains '-_' instead of '+/'.
        
        Returns:
            str: base64 encoded.
        
        """
        # Extract the bits
        bits = self.extract_bitarray(positions=positions)
        
        # Convert to base64
        return bitarray_to_base64(bits, url_safe=url_safe)
    
    
    def insert(self, b, positions):
        """Insert another BitCollection into this collection.
        
        Args:
            b (BitCollection): Content to insert.
            positions (list): The positions to insert bits into the
                collection.
        
        """
        self.insert_bitarray(b.content, positions=positions)
    
    
    def insert_bitarray(self, b, positions):
        """Insert bits into this collection at the given positions.
        
        Args:
            b (bitarray): Content to insert.
            positions (list): The positions to insert bits into the
                collection.
        
        """
        # Roll through each position.
        for j, position in enumerate(positions):
            
            # Put the value into this collection
            self.content.insert(position, b[j])
    
    
    def insert_bytes(self, b, positions):
        """Insert byte bits into this collection at the given positions.
        
        Args:
            b (bytes|bytearray): Content to insert.
            positions (list): The positions to insert bits into the
                collection.
        
        """
        # Use the bitarray
        bits = bitarray()
        bits.frombytes(b)
        
        # Run the bitarray function
        self.insert_bitarray(bits, positions=positions)
    
    
    def insert_int(self, i, positions):
        """Insert the bits of the integer into this collection.
        
        Args:
            i (int): Content to insert.
            positions (list): The positions to insert bits into the
                collection.
        
        """
        # Roll through each position.
        for j, position in enumerate(positions):
            
            # Get the bit from the integer
            bit = (i & (1 << j)) >> j
            
            # Insert it into the content
            self.content.insert(position, bit)
    
    
    def insert_hex(self, s, positions):
        """Insert a hexadecimal string into this collection.
        
        Args:
            s (str): Hexadecimal string to insert.
            positions (list): The positions to insert bits into the
                collection.
        
        """
        self.insert_bitarray(hex_to_bitarray(s), positions=positions)
    
    
    def insert_base64(self, s, positions, url_safe=False):
        """Insert byte bits into this collection at the given positions.
        
        Args:
            s (str): Base64 string to insert.
            positions (list): The positions to insert bits into the
                collection.
            url_safe (Optional[bool]): Whether the input string contains
                '-_' instead of '+/'.
        
        """
        bits = base64_to_bitarray(s, url_safe=url_safe)
        self.insert_bitarray(bits, positions=positions)
    
    
    def length(self):
        """Return the number of bits stored in the object."""
        if not self.content:
            return 0
        return self.content.length()
    
    
    def pop(self, index=None):
        """Pop the last value off the bitarray."""
        if not index:
            index = self.length() - 1
        return self.content.pop(index)
    
    
    def to_base64(self, url_safe=False):
        """Express this collection as a base64 string.
        
        Args:
            url_safe (Optional[bool]): Whether the output should contain
                '-_' instead of '+/'.
        
        Returns:
            str: base64 encoded.
        
        """
        return bitarray_to_base64(self.content, url_safe=url_safe)
    
    
    def to_bytes(self):
        """Express this collection as bytes."""
        return self.content.tobytes()
    
    
    def to_hex(self):
        """Express this collection as a hexadecimal string."""
        return bitarray_to_hex(self.content)
    
    
    def to_int(self):
        """Express this collection as an integer."""
        return bitarray_to_int(self.content)


class SecretKeyCollection:
    """Standard object to ingest and express a secret key.
    
    Secret keys are provided as strings that will be decoded as ASCII
    characters. Values will be restricted to characters 0x20 thru 0x7F
    and reorganized as compatible bytes in order to maximize entropy.
    """
    
    def __init__(self, s):
        """Initilize a collection with a single secret key."""
        
        # Convert string into bytes according to the ASCII standard.
        bytes_ = bytes(s, 'ascii')
        
        # Loop through bytes and squash into smaller amount of bits.
        bits = bitarray()
        for b in bytes_:
            t = (int(b) - 32) & 0x5F
            bits = int_to_bitarray(t,6) + bits
        
        # Pad if necessary.
        mod = bits.length() % 8
        if mod:
            bits = bitarray('0' * (8 - mod)) + bits
        
        # Keep the data.
        self.original = s
        self.content = bits
        self.index = 0
    
    
    def chunk(self, n):
        """Spit out even-sized chunks from the secret key bits.
        
        Args:
            n (int): Number of evenly-sized chunks.
            
        Returns:
            list: ints wrapped at 32 bits.
        
        """
        # How big are chunks (in bytes)?
        total_bytes = self.content.length() // 8
        chunk_size = total_bytes // n
        
        # Make sure all characters are used (if size was floored)
        if total_bytes % n:
            chunk_size += 1
        
        # Get the iterator going
        iterator = iter(self.content.tobytes())
        for a in range(n):
            
            # Build this specific chunk, one char at a time
            bytes_ = bytearray()
            for b in range(chunk_size):
                
                # Add the character to the chunk
                try:
                    bytes_.append(next(iterator))
                
                # String stopped early
                except StopIteration:
                    break
            
            # Hash the bytes.
            m = hashlib.sha256()
            m.update(bytes_)
            d = m.digest()
            
            # Chunk completed, yield the 32 bits as an int.
            yield int.from_bytes(d[0:4], byteorder='big')
            



import base64
import binascii
from bitarray import bitarray


def bitarray_to_base64(b, url_safe=False):
    """Convert a bitarray to a base64 encoded string."""
    # Get the bytes
    bytes_ = b.tobytes()
    
    # Convert it to a string
    string = binascii.b2a_base64(bytes_).decode('ascii').rstrip('\n')
    mod = len(string) % 3
    if mod:
        string += '=' * (3 - mod)
    
    # Make it url-safe
    if url_safe:
        string = string.replace('+','-').replace('/','_')
    
    # Return the string
    return string


def bitarray_to_hex(b):
    """Convert a bitarray to a hexidecimal string."""
    bytes_ = b.tobytes()
    s = binascii.hexlify(bytes_).decode('ascii')
    if b.length() % 8 and not (b.length() % 4):
        return s[:-1]
    return s


def bitarray_to_bytes(b):
    """Convert a bitarray to bytes."""
    return b.tobytes()


def bitarray_to_int(b):
    """Convert a bitarray to an integer."""
    if not len(b):
        return 0
    return int(b.to01(), 2)


def bitarray_to_str(b, codec):
    """Convert at bitarray to a string.
    
    Args:
        codec (str): How to decode the bits. Examples include 'ascii'
            and 'utf-8'.
    
    Returns:
        str: made from the decoded bytes from the bits.
    
    """
    return b.tobytes().decode(codec)


def base64_to_bitarray(s, url_safe=False):
    """Convert a base64 string to a bitarray.
    
    Args:
        s (str): Base64 encoded string.
        url_safe (bool): Whether to substitute '-_' with '+/'.
    
    Returns:
        bitarray: made from the base64 string.
    
    """
    # First, make sure the b64 is properly padded and formatted
    if url_safe:
        s = s.replace('-','+').replace('_','/')
    s += '=='
    
    # Decode it
    bytes_ = base64.b64decode(s)
    a = bitarray()
    a.frombytes(bytes_)
    return a


def bytes_to_bitarray(b):
    """Convert bytes into a bitarray."""
    a = bitarray()
    a.frombytes(b)
    return a


def int_to_bitarray(i, bits):
    """Convert integer to a bitarray.
    
    Args:
        i (int): Value to convert.
        bits (int): Number of bits for this integer.
    
    Returns:
        str: binary representation of the integer.
    
    """
    output = bitarray()
    for j in range(0,bits):
        bit = (i & (1 << j)) >> j
        output.insert(0, bit)
    return output
    

def hex_to_bitarray(s):
    """Convert hexidecimal string into a bitarray."""
    padded = False
    if len(s) % 2:
        padded = True
        s += '0'
    bytes_ = binascii.unhexlify(s)
    a = bitarray()
    a.frombytes(bytes_)
    if padded:
        for i in range(4):
            a.pop()
    return a


def int_to_binstr(i, bits):
    """Convert integer to a '01' string.
    
    Args:
        bits (int): Number of bits for this integer.
    
    Returns:
        str: binary representation of the integer.
    
    """
    output = ''
    for j in range(0,bits):
        bit = (i & (1 << j)) >> j
        output = str(bit) + output
    return output


def str_to_bitarray(s, codec):
    """Convert a string to a bitarray.
    
    Args:
        codec (str): How to encode the string into bits. Examples
            include 'ascii' and 'utf-8'.
    
    Returns:
        bitarray: made from the input string.
    
    """
    return bitarray().frombytes(s.encode(codec))


def insert_bits(source, insert, positions):
    """Distribute bits from the inserted value into the source.
    
    This function will insert each bit sequentially, meaning that as
    bits are inserted, they will offset the position of later bits.
    The result will affect the position of all existing and added bits
    on each iteration.
    
    Args:
        source (int): Data in integer form to have bits spliced in.
        insert (int): Data to put into the source.
        positions (list): Ordered integers for splicing data into the
            source. Each number is iterated through in order.
        
    Returns:
        Integer representing the resulting data.
    
    """
    # Figure out the length for managing the top bit shifting
    length = source.bit_length()
    
    # Each position also has an index (for the insert)
    for i, position in enumerate(positions):
        
        # Make sure the string is always as long as the positions
        length = max(length, position)
        
        # Create the top and bottom buns
        bottom_mask = 2**position - 1
        top_mask = (2**(length + i + 1) - 1) ^ bottom_mask
        
        # Get the bit we'll insert
        insert_bit = (2**i & insert) >> i
        
        # Open up the buns and insert the burger
        source = ((top_mask & source) << 1) | (source & bottom_mask)
        source |= insert_bit << (position)
        
    # Return the full quarter-pounder
    return source


def extract_bits(source, positions):
    """Get information from the source data based on bit positions.
    
    This function will extract each bit sequentially, meaning that as
    bits are removed, they will offset the position of later bits. The
    result will affect the osition of all existing and added bits on
    each iteration.
    
    Args:
        source (int): Data to extract information from.
        positions (list): Ordered integers indicating the locations of
            bits to extract. Extraction will cascade and affect the
            location of later bits.
        
    Returns:
        source (int): Source after extraction.
        extracted (int): Data that was extracted.
    
    """
    # Figure out the length for managing the bit shifting
    length = source.bit_length()
    positions_length = len(positions) - 1
    
    # Get the result ready
    extracted = 0
    
    # Each position also has an index (for the insert)
    for i, position in enumerate(positions):
        
        # Create the top and bottom buns
        bottom_mask = 2**position - 1
        top_mask = ((2**(length - i + 1) - 1) - 2**position) ^ bottom_mask
        
        # Extract a patty
        prize_bit = (2**position & source) >> position
        extracted |= prize_bit << (positions_length - i)
        
        # Collapse the buns without the burger
        source = ((top_mask & source) >> 1) | (source & bottom_mask)
        
    # Return the buns and patties seperately
    return source, extracted
    


import binascii
import copy
import hashlib

from .collections import BitCollection, SecretKeyCollection
from .exceptions import ConfigError
from .random import MT19937


class TokenLayer:
    """Provides a common interface for several token layer types."""
    
    TYPES = [
        'BitCollection',
        'int',
        'bytes',
        'hex',
    ]
    """The list of valid types."""
    
    def __init__(self, d):
        """Takes a dictionary of settings and ingests it as a layer."""
        
        # Must be a dictionary.
        if not isinstance(d, dict):
            raise ConfigError('layers must each be a dict')
        self.config = d
        
        # What is this layer's type?
        self.type = d.get('type', None)
        if self.type not in self.TYPES:
            err = 'layer type must be in %s' % ', '.join(self.TYPES)
            raise ConfigError(err)
        
        # Init the bits.
        self.bits = d.get('bits', None)
        if self.bits != None and not isinstance(self.bits, int):
            raise ConfigError('layer bits key must be a positive int')
        
        # Init the length.
        self.length = d.get('length', None)
        if not self.length and not self.bits:
            raise ConfigError('layer bits key must be a positive int')
        if self.length:
            if not isinstance(self.length, int) or self.length < 1:
                raise ConfigError('layer length must be a positive int')
        
        # Bits and ints require bits.
        if self.type in ['BitCollection','int']:
            if not self.bits:
                err = 'layer %s bits key must be a positive int'
                raise ConfigError(err % self.type)
            self.length = self.bits
        
        # Bytes are 8 bits each.
        elif self.type == 'bytes':
            if self.bits:
                if (self.bits % 8) != 0:
                    raise ConfigError('layer bytes bits must be divisible by 0')
                self.length = self.bits // 8
            else:
                self.bits = self.length * 8
        
        # Hex are 16 bits per character.
        elif self.type == 'hex':
            if self.bits:
                if (self.bits % 4) != 0:
                    raise ConfigError('layer hex bits must be divisible by 16')
                self.length = self.bits // 4
            else:
                self.bits = self.length * 4
        
        # Type doesn't count.
        else:
            raise ConfigError('invalid layer type')
        
        # Every type has the optional 'positions' key
        self.positions = None
        positions = d.get('positions', None)
        if positions:
            
            # Make sure the positions are syntacticly valid
            err = 'positions must be a list of non-negative ints'
            if not isinstance(positions, list):
                raise ConfigError(err)
            for value in positions:
                if not isinstance(value, int) or value < 0:
                    raise ConfigError(err)
            
            # Do the number of positions match the number of bits?
            if len(positions) != self.bits:
                raise ConfigError('number of layer positions must match bits')
            
            # Positions is valid
            self.positions = positions
    
    
    def to_bitcollection(self, v):
        """Get the BitCollection for this layer."""
        
        # Is it a BitCollection?
        if self.type == 'BitCollection':
            if not isinstance(v, BitCollection):
                raise ValueError('layer value must be a BitCollection')
            if v.length() != self.bits:
                raise ValueError('layer value has incorrect number of bits')
            return v
        
        # Is it an int?
        if self.type == 'int':
            if not isinstance(v, int):
                raise ValueError('layer value must be an int')
            if v.bit_length() > self.bits:
                raise ValueError('layer value is too many bits')
            return BitCollection.from_int(v, bits=self.bits)
        
        # Is it bytes?
        if self.type == 'bytes':
            if not isinstance(v, bytes):
                raise ValueError('layer value must be bytes')
            if len(v) != self.length:
                raise ValueError('layer value is incorrect length')
            return BitCollection.from_bytes(v)
        
        # Is it string?
        if self.type == 'hex':
            if not isinstance(v, str):
                raise ValueError('layer value must be str')
            if len(v) != self.length:
                raise ValueError('layer value is incorrect length')
            return BitCollection.from_hex(v)
        
        # Something failed here, which should be impossible.
        raise ConfigError('unable to create BitCollection')
    
    
    def from_bitcollection(self, b):
        """Convert to original format."""
        if self.type == 'BitCollection':
            return b
        if self.type == 'int':
            return b.to_int()
        if self.type == 'bytes':
            return b.to_bytes()
        if self.type == 'hex':
            return b.to_hex()
        raise ConfigError('unable to convert to original format')


class TokenResult:
    """Object created from decoding a token.
    
    The Token object only carries configuration information. It's
    resulting collection carries its results.
    """
    
    def __init__(self, private_token=None, public_token=None, layers=None):
        """Populate the collection once with everything it'll ever have.
        
        """
        self.layers = layers
        self.public_token = public_token
        self.private_token = private_token


class Token:
    """Generate and encode tokens based on ordered parameters. 
    
    TokenGenerator methods affect how the final token will be created.
    Some methods overwrite their previous values, while others append
    to ordered lists.
    
    A token option can optionally be given a single config dictionary
    with all the necessary parameters. This is an easier method of
    sharing token configurations between functions and modules.
    
    A sample config could be stored as and submitted as the following:
        config = {
            "secret_key": "the length of this should be long",
            "random_bits": 512,
            "seed_bits": 4,
            "layers": [
                {
                    'type': 'int',
                    'bits': 64, # 64 bits
                },
                {
                    'type': 'int',
                    'bits': 5,
                    'positions': [1,5,2,3,5], # Manual bit positions
                },
            ],
        }
    
    Here is a sample usage:
        >>> token = Token(config)
        >>> result1 = token.encode(1,2)
        >>> result2 = token.decode(result1.public_token)
        >>> result2.layers[1] == 2
        True
    
    """
    
    def __init__(self, config=None):
        """Set default object properties."""
        
        # The secret string to use for predictable randomness.
        self.secret_key = None
        
        # Number of bits saying how many automatic positions exist.
        self.seed_bits = 8
        
        # What should the random token look like?
        self.stored_token_bits = 512
        
        # Values getting spliced into the public token.
        self.layers = []
        
        # Is config here?
        self.config = {}
        if config:
            self.set_config(config)
        
    
    def set_config(self, config):
        """Ingest a config dictionary.
        
        Raises:
            ConfigError: layers must be lists or ints.
        
        """
        # Keep this config.
        self.config = config
        
        # Ingest the secret used to seed everything.
        secret = config.get('secret_key', None)
        if secret:
            if not isinstance(secret, str):
                raise ConfigError('secret key must be a string')
            self.secret_key = secret
        
        # Might have to use global secret if available
        else:
            from token_cloak import secret_key
            if not secret_key:
                raise ConfigError('secret key is not set')
            self.secret_key = secret_key
        
        # Get the secret key collection going.
        self.secret_key_collection = SecretKeyCollection(self.secret_key)
        
        # Determines the length of stored token.
        self.stored_token_bits = 0
        random_bits = config.get('random_bits', None)
        if random_bits:
            
            # Is it a direct injection?
            if isinstance(random_bits, BitCollection):
                self.stored_token_bits = random_bits
            
            # Is it just a length for random generation?
            else:
                if not isinstance(random_bits, int) or random_bits < 0:
                    raise ConfigError('random bits must be a non-negative int')
                
                # Generate the token now
                self.stored_token_bits = random_bits
        
        # Determines how many possible seeds there can be.
        seed_bits = config.get('seed_bits', None)
        if seed_bits:
            if not isinstance(seed_bits, int) or seed_bits < 0:
                raise ConfigError('seed bits must be a non-negative int')
            self.seed_bits = seed_bits
        
        # Ingest the layers sequence and sizes.
        self.layers = []
        if config.get('layers', None):
            if not isinstance(config.get('layers'), list):
                raise ConfigError('layers must be a list')
            
            # Ingest layers as TokenLayers.
            for row in config.get('layers'):
                self.layers.append(TokenLayer(row)) # Raises ConfigError
        
        # Make sure the secret is long enough for layers.
        if len(self.layers) > len(self.secret_key):
            err = "secret key length cannot be less than number of layers"
            raise ConfigError(err)
    
    
    def encode(self, *args):
        """Make the public token based on the input values.
        
        If arguments are supplied with this method, then they will
        override any current settings on the object.
        
        Returns:
            BitCollection: public token.
        
        Raises:
            ConfigError: number of args doesn't match number of layers.
        
        """
        # Ensure the input matches
        stored_token = None
        if len(args) != len(self.layers):
            
            # Is there not just one more than before?
            if len(args) != (len(self.layers) + 1):
                err = 'number of args doesn\'t match number of layers'
                raise ConfigError(err)
            
            # The first argument will be treated as a token.
            stored_token = args[0]
            if not isinstance(args[0], BitCollection):
                raise ValueError('token must be a BitCollection')
            
            # Is it the correct length?
            if stored_token.length() != self.stored_token_bits:
                err = 'token must be %d bits long, is currently %d'
                err = err % (self.stored_token_bits, stored_token.length(),)
                raise ValueError(err)
            
            # Adjust the args for proper use.
            args = tuple(list(args)[1:])
        
        # Generate a new stored token.
        if not stored_token:
            if self.stored_token_bits and self.stored_token_bits > 0:
                stored_token = BitCollection.from_random(self.stored_token_bits)
            else:
                stored_token = BitCollection()
        
        # Start the new public token.
        public_token = copy.deepcopy(stored_token)
        
        # Are there any layers?
        if not self.layers:
            return TokenResult(
                    public_token=public_token,
                    private_token=stored_token)
        
        # Decide on predictable seed positions up front
        seed_sources = []
        need_seeds = self.needed_seeds()
        if need_seeds > 0:
            for chunk in self.secret_key_collection.chunk(need_seeds):
                seed_sources.append(chunk)
            seed_sources = seed_sources[::-1]
        
        # Go through each layer in order
        stored_layers = []
        for index, layer in enumerate(self.layers):
            
            # Does this layer have positions?
            layer_seed_seed = None
            layer_seed_value = 0
            layer_positions = layer.positions
            if not layer_positions:
                
                # Get the seed prepared.
                layer_seed_value = seed_sources.pop()
                if self.seed_bits:
                    layer_seed_seed = layer_seed_value
                    b = BitCollection.from_random(self.seed_bits)
                    layer_seed_value = b.to_int()
                    
                # Generate the layer positions using the seed.
                layer_positions = self.generate_bit_positions(
                        seed=layer_seed_value,
                        max_position=public_token.length(),
                        bits=layer.bits)
            
            # Sew in the new bits.
            public_token.insert(
                    layer.to_bitcollection(
                            args[index]), positions=layer_positions)
            
            # Save the content.
            stored_layers.append(args[index])
            
            # Was there an automatic seed?
            if layer_seed_seed:
                
                # Generate the seed positions.
                seed_positions = self.generate_bit_positions(
                        seed=layer_seed_seed,
                        max_position=public_token.length(),
                        bits=self.seed_bits)
                
                # Sew in the seed bits.
                public_token.insert_int(
                        layer_seed_value, positions=seed_positions)
        
        # All spliced - return results.
        return TokenResult(
                public_token=public_token,
                private_token=stored_token,
                layers=stored_layers)
    
    
    def decode(self, token, data_type=None, **kwargs):
        """Decode a token created by this class.
        
        For accurate decoding, it is essential that the input
        parameters used to create the token are provided in the same
        manner as they were for generation.
        
        The main difference is the 'values' attribute will be updated
        with the decoded values from the token.
        
        Args:
            token (mixed): public token in a variety of possible types.
            data_type (Optional[str]): How the token is encoded on a
                data level. Optional if not string or if in config.
        
        Returns:
            If successful, dict. Otherwise, None.
        
        """
        # Get data_type from somewhere else.
        if not data_type:
            data_type = self.config.get('public_token_type', None)
        
        # Start with the expected length.
        expected_length = self.public_token_bit_length()
        bit_remainder = 0
        
        # Put the token into a BitCollection based on data_type.
        if not data_type:
            if not isinstance(token, BitCollection):
                raise ValueError('token\'s data type does not match')
            public_token = token
        
        # Decode from base64.
        elif data_type == 'base64':
            try:
                public_token = BitCollection.from_base64(
                        token, url_safe=kwargs.get('url_safe', None))
            except binascii.Error:
                return None
            bit_remainder = self.remainder_by_divisor(expected_length, 8)
        
        # Decode from bytes.
        elif data_type == 'bytes':
            public_token = BitCollection.from_bytes(token)
            bit_remainder = self.remainder_by_divisor(expected_length, 8)
        
        # Decode from hex.
        elif data_type == 'hex':
            public_token = BitCollection.from_hex(token)
            bit_remainder = self.remainder_by_divisor(expected_length, 4)
        
        # Decode from int.
        elif data_type == 'int':
            if token.bit_length() > self.public_token_bit_length():
                return None
            public_token = BitCollection.from_int(
                    token, bits=self.public_token_bit_length())
        
        # Invalid type.
        else:
            raise ValueError('invalid data_type')
        
        # Adjust if a remainder exists.
        for i in range(bit_remainder):
            public_token.pop()
        
        # Validate the token by its length.
        if expected_length != public_token.length():
            return None
        
        # Setup the stored token.
        stored_token = copy.deepcopy(public_token)
        
        # Are there layers?
        if not self.layers:
            return TokenResult(
                    public_token=public_token,
                    private_token=stored_token)
        
        # Setup for peeling away layers.
        stored_layers = []
        seed_sources = []
        need_seeds = self.needed_seeds()
        if need_seeds > 0:
            for chunk in self.secret_key_collection.chunk(need_seeds):
                seed_sources.append(chunk)
        
        # Start off with the layers!
        for index, layer in enumerate(self.layers[::-1]):
            
            # Does it have a seeded position?
            layer_seed_seed = None
            layer_seed_value = 0
            layer_positions = layer.positions
            if not layer_positions:
                
                # Get the seed from the token.
                layer_seed_value = seed_sources.pop()
                if self.seed_bits:
                    seed_positions = self.generate_bit_positions(
                            seed=layer_seed_value,
                            max_position=stored_token.length() - self.seed_bits,
                            bits=self.seed_bits)
                    
                    # Extract in the seed bits.
                    layer_seed_value = stored_token.extract_int(
                            positions=seed_positions[::-1])
                
                # Generate the layer positions using the seed.
                layer_positions = self.generate_bit_positions(
                        seed=layer_seed_value,
                        max_position=stored_token.length() - layer.bits,
                        bits=layer.bits)
                
            # Get the layer value from the token based on format.
            layer_value = stored_token.extract(
                    positions=layer_positions[::-1])
            
            # Store the value away as its original datatype.
            stored_layers.append(layer.from_bitcollection(layer_value))
        
        # Reverse the stored_layers in order to match how layers are added.
        stored_layers = stored_layers[::-1]
        
        # All done!
        return TokenResult(
                public_token=public_token,
                private_token=stored_token,
                layers=stored_layers)
    
    
    def hash_seed(self, seed):
        """
        Takes a seed and returns another seed that's been hashed with
        this token's secret key.
        
        Args:
            seed (int): Randomly generated seed value.
            
        Returns:
            int: hashed with this token's secret key.
        
        """
        # Start up our SHA256 object.
        m = hashlib.sha256()
        
        # Insert the bytes from our original seed.
        b = BitCollection.from_int(seed, bits=self.seed_bits)
        m.update(b.to_bytes())
        
        # Insert the bytes from the secret key.
        m.update(self.secret_key.encode('ascii'))
        
        # Get 32-bits worth of the resulting hash.
        c = BitCollection.from_bytes(m.digest()[:4])
        return c.to_int()
    
    
    def needed_seeds(self):
        """Calculates number of layer seeds needed to be generated."""
        need_seeds = 0
        for layer in self.layers:
            if not layer.positions:
                need_seeds += 1
        return need_seeds
    
    
    def public_token_bit_length(self):
        """
        Helper function to calculate supposed bit length for token.
        
        Args:
            with_padding (Optional[str]): Data type to pad for.
        
        Returns:
            int: number of bits expected for the public token.
        
        """
        # How many random bits are there?
        total_bits = self.stored_token_bits
        
        # Do the math for each layer
        if self.layers:
            for layer in self.layers:
                
                # Add bits for included seed
                if not layer.positions:
                    total_bits += self.seed_bits
                
                # Add the number of bits the actual value can be
                total_bits += layer.bits
        
        # Return the tally
        return total_bits
        
    
    @staticmethod
    def remainder_by_divisor(expected, divisor):
        """Shortcut to get the remainder from the expected length.
        
        Since several input formats are chunked in even, multi-bit
        blocks, sometimes ending bits need to be sliced off in order to
        reproduce the original token.
        
        The resulting remainder is that number of bits.
        
        Args:
            expected (int): Length of expected public token.
            divisor (int): Multible by which the input data format is
                chunked by.
        
        Returns:
            int: from 0 to divisor - 1.
        
        """
        mod = expected % divisor
        if mod:
            return divisor - mod
        return 0
    
    
    def generate_bit_positions(self, seed, max_position, bits):
        """
        Generates an ordered list of integer positions based on the
        provided elements.
        
        Args:
            seed (mixed): Hashable value to create predictable randomness.
            max_position (int): Highest allowed position to generate.
            bits (int): Number of positions needed for this operation.
        
        Returns:
            positions (list): List of integers, denoting the position in
                the public token for bits to reside.
        
        """
        # Seed the randomness according to the seed
        r = MT19937(self.hash_seed(seed))
        
        # Start generating
        positions = []
        for i in range(bits):
            positions.append(r.rand_int(0, max_position + i))
        
        # Return the list of integers
        return positions
    


import os
import os.path
import re
from setuptools import setup


def get_info(var):
    """Get version from the package."""
    with open(os.path.join('token_cloak','__init__.py')) as f:
        content = f.read()
    return re.search(var + r'\s*=\s*["\'](.+?)["\']', content).group(1)


def get_packages(package):
    """
    Taken from https://github.com/tomchristie/django-rest-framework/blob/master/setup.py
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Taken from https://github.com/tomchristie/django-rest-framework/blob/master/setup.py
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


VERSION = get_info('__version__')
LICENSE = get_info('__license__')


setup(
        name="token_cloak",
        description="A utility to hide data in public tokens.",
        url="https://github.com/ryannjohnson/token-cloak-python",
        license=LICENSE,
        version=VERSION,
        packages=get_packages('token_cloak'),
        install_requires=[
            'bitarray',
        ],
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
        ])


"""
Token Cloak is a utility for lacing public tokens with data.
"""

__license__ = 'MIT License'
__version__ = '0.1.0'


secret_key = None
"""Controls pseudo-randomness within Token Cloak."""

# Token
from .tokens import Token

# BitCollection
from .collections import BitCollection